/*******************************************************************
                 lanczos 3-lobed resampling filter
 *******************************************************************/
#include <windows.h>
#include <zmouse.h>
#include <math.h>
#include <float.h>
#include "filter.h"

#define MIN_WIDTH   32
#define MIN_HEIGHT  32
#define MAX_WIDTH  856
#define MAX_HEIGHT 576

#ifndef PI
#define PI (atan(1)*4)
#endif

#define ID_BUTTON_640X480 40001
#define ID_BUTTON_352X240 40002
#define ID_BUTTON_320X240 40003
#define ID_BUTTON_SETSIZE 40004

#define ID_BUTTON_NO_SIMD 40005
#define ID_BUTTON_SSE2INT 40006
#define ID_BUTTON_SSE_FLT 40007
#define ID_BUTTON_SSE2DBL 40008
#define ID_BUTTON_MMX_INT 40009

/*------------------------------------------------------------------
   global variables
 ------------------------------------------------------------------*/
typedef struct {
	HFONT font;
	HWND  simd_group;
	HWND  rb_no_simd;
	HWND  rb_sse2_int;
	HWND  rb_sse_float;
	HWND  rb_sse2_double;
	HWND  rb_mmx_int;
	HWND  old_simd;
	HWND  b640x480;
	HWND  b352x240;
	HWND  b320x240;
	HWND  b_set_size;
	HWND  e_width;
	HWND  e_height;
	HWND  label;
} FILTER_DIALOG;

FILTER_DIALOG dialog;

typedef struct {
	int   length;
	int   tap;
	int   width;
	int   height;
	int   step;
	int **weight;
	int **index;
} PARAMETER;

PARAMETER x_prm;
PARAMETER y_prm;

typedef struct {
	int     length;
	int    *index;
	float **weight;
	float **alignment_weight;
	int     x_loop_count;
	int     y_loop_count;
	int     tap;
	int     step;
	int     src_length_minus_one;
	int     center_loop_count;
} SSE_PARAMETER;

SSE_PARAMETER x_sse_prm;
SSE_PARAMETER y_sse_prm;

typedef struct {
	int     length;
	int    *index;
	int   **weight;
	int   **alignment_weight;
	int     x_loop_count;
	int     y_loop_count;
	int     tap;
	int     step;
	int     src_length_minus_one;
	int     center_loop_count;
} SSE2_INT_PARAMETER;

SSE2_INT_PARAMETER x_sse2_int_prm;
SSE2_INT_PARAMETER y_sse2_int_prm;

typedef struct {
	int      length;
	int     *index;
	double **weight;
	double **alignment_weight;
	int      x_loop_count;
	int      y_loop_count;
	int      tap;
	int      step;
	int      src_length_minus_one;
	int      center_loop_count;
} SSE2_DOUBLE_PARAMETER;
	
SSE2_DOUBLE_PARAMETER x_sse2_double_prm;
SSE2_DOUBLE_PARAMETER y_sse2_double_prm;

typedef struct {
	int     length;
	int    *index;
	int   **weight;
	int   **alignment_weight;
	int     x_loop_count;
	int     y_loop_count;
	int     tap;
	int     step;
	int     src_length_minus_one;
	int     center_loop_count;
} MMX_INT_PARAMETER;

MMX_INT_PARAMETER x_mmx_int_prm;
MMX_INT_PARAMETER y_mmx_int_prm;

int    src_width;
int    src_height;
int    rst_width;
int    rst_height;

int    simd;

float  sse_half[4];
double sse2_half[2];

WNDPROC default_button_window_procedure;

/*------------------------------------------------------------------
   function prototypes
 ------------------------------------------------------------------*/
BOOL func_proc_sse2_int( FILTER *fp,FILTER_PROC_INFO *fpip );
BOOL func_proc_sse_float( FILTER *fp,FILTER_PROC_INFO *fpip );
BOOL func_proc_sse2_double( FILTER *fp,FILTER_PROC_INFO *fpip );
BOOL func_proc_mmx_int( FILTER *fp,FILTER_PROC_INFO *fpip );

static void proc_horz_nosimd(int thread_id, int thread_num, void *param1, void *param2);
static void proc_vert_nosimd(int thread_id, int thread_num, void *param1, void *param2);
static void proc_horz_sse2_int(int thread_id, int thread_num, void *param1, void *param2);
static void proc_vert_sse2_int(int thread_id, int thread_num, void *param1, void *param2);
static void proc_horz_sse_float(int thread_id, int thread_num, void *param1, void *param2);
static void proc_vert_sse_float(int thread_id, int thread_num, void *param1, void *param2);
static void proc_horz_sse2_double(int thread_id, int thread_num, void *param1, void *param2);
static void proc_vert_sse2_double(int thread_id, int thread_num, void *param1, void *param2);
static void proc_horz_mmx_int(int thread_id, int thread_num, void *param1, void *param2);
static void proc_vert_mmx_int(int thread_id, int thread_num, void *param1, void *param2);

static int    is_setup_required(FILTER *fp, FILTER_PROC_INFO *fpip);

static void   setup_table(FILTER *fp, FILTER_PROC_INFO *fpip);
static void   setup_sse_table(FILTER *fp, FILTER_PROC_INFO *fpip);
static void   setup_sse2_int_table(FILTER *fp, FILTER_PROC_INFO *fpip);
static void   setup_sse2_double_table(FILTER *fp, FILTER_PROC_INFO *fpip);
static void   setup_mmx_int_table(FILTER *fp, FILTER_PROC_INFO *fpip);

static void   setup_interpolation_parameter(int source_length, int result_length, PARAMETER *out);
static void   setup_sse_interpolation_parameter(int source_length, int result_length, SSE_PARAMETER *out);
static void   setup_sse2_int_interpolation_parameter(int source_length, int result_length, SSE2_INT_PARAMETER *out);
static void   setup_sse2_double_interpolation_parameter(int source_length, int result_length, SSE2_DOUBLE_PARAMETER *out);
static void   setup_mmx_int_interpolation_parameter(int source_length, int result_length, MMX_INT_PARAMETER *out);

static void   setup_decimation_parameter(int source_length, int result_rength, PARAMETER *out);
static void   setup_sse_decimation_parameter(int source_length, int result_rength, SSE_PARAMETER *out);
static void   setup_sse2_int_decimation_parameter(int source_length, int result_rength, SSE2_INT_PARAMETER *out);
static void   setup_sse2_double_decimation_parameter(int source_length, int result_rength, SSE2_DOUBLE_PARAMETER *out);
static void   setup_mmx_int_decimation_parameter(int source_length, int result_rength, MMX_INT_PARAMETER *out);

static void   clear_parameter(PARAMETER *prm);
static void   clear_sse_parameter(SSE_PARAMETER *prm);
static void   clear_sse2_int_parameter(SSE2_INT_PARAMETER *prm);
static void   clear_sse2_double_parameter(SSE2_DOUBLE_PARAMETER *prm);
static void   clear_mmx_int_parameter(MMX_INT_PARAMETER *prm);

static double lanczos3_weight(double phase);
static void   horizontal_filter(PIXEL_YC *in, PIXEL_YC *out, PARAMETER *prm);
static void   vertical_filter(PIXEL_YC *in, PIXEL_YC *out, PARAMETER *prm);
static int    is_mmx_enable();
static int    is_sse_enable();
static int    is_sse2_enable();

static void   init_dialog(HWND hwnd, HINSTANCE hinst);
static void   set_subclass();
static void   remove_subclass();
static void   enable_simd_radio_button();
static void   load_ini_value(FILTER *fp);
static void   save_ini_value(FILTER *fp);

static void   on_no_simd_button(FILTER *fp);
static void   on_sse2_int_button(FILTER *fp);
static void   on_sse_float_button(FILTER *fp);
static void   on_sse2_double_button(FILTER *fp);
static void   on_mmx_int_button(FILTER *fp);

static void   on_640x480_button(FILTER *fp);
static void   on_352x240_button(FILTER *fp);
static void   on_320x240_button(FILTER *fp);
static void   on_setsize_button(FILTER *fp);

LRESULT CALLBACK sse_button_window_procedure(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);
LRESULT CALLBACK sse2_button_window_procedure(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);
LRESULT CALLBACK mmx_button_window_procedure(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);
LRESULT CALLBACK disable_button_window_procedure(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);

static void   setup_range(FILTER *fp, void *editp);

/*------------------------------------------------------------------
   assembler functions
 ------------------------------------------------------------------*/
extern void __stdcall horizontal_filter_sse_float(PIXEL_YC *in, PIXEL_YC *out, int *ref_left, float **weight, int width, int height, int tap, int step, int src_width_minus_one, int center_width);
extern void __stdcall vertical_filter_sse_float(PIXEL_YC *in, PIXEL_YC *out, int *ref_top, float **weight, int x_loop_count, int height, int tap, int step, int src_height_minus_one, int center_height);
extern void __stdcall horizontal_filter_sse2_int(PIXEL_YC *in, PIXEL_YC *out, int *ref_left, int **weight, int width, int height, int tap, int step, int src_width_minus_one, int center_width);
extern void __stdcall vertical_filter_sse2_int(PIXEL_YC *in, PIXEL_YC *out, int *ref_top, int **weight, int x_loop_count, int height, int tap, int step, int src_height_minus_one, int center_height);
extern void __stdcall horizontal_filter_sse2_double(PIXEL_YC *in, PIXEL_YC *out, int *ref_left, double **weight, int width, int height, int tap, int step, int src_width_minus_one, int center_width);
extern void __stdcall vertical_filter_sse2_double(PIXEL_YC *in, PIXEL_YC *out, int *ref_top, double **weight, int x_loop_count, int height, int tap, int step, int src_height_minus_one, int center_height);
extern void __stdcall horizontal_filter_mmx_int(PIXEL_YC *in, PIXEL_YC *out, int *ref_left, int **weight, int width, int height, int tap, int step, int src_width_minus_one, int center_width);
extern void __stdcall vertical_filter_mmx_int(PIXEL_YC *in, PIXEL_YC *out, int *ref_top, int **weight, int x_loop_count, int height, int tap, int step, int src_height_minus_one, int center_height);

/*------------------------------------------------------------------
   filter information
 ------------------------------------------------------------------*/
#define TRACK_N 2
char  *track_name[] =    {"X", "Y",};
int    track_default[] = {320,240,};
int    track_s[] =       {MIN_WIDTH,MIN_HEIGHT};
int    track_e[] =       {MAX_WIDTH,MAX_HEIGHT};

#define CHECK_N 0
char  *check_name[] =    {"", };
int    check_default[] = {0, };

FILTER_DLL filter = {
	FILTER_FLAG_INFORMATION|FILTER_FLAG_WINDOW_SIZE,
	320,176,
	"Lanczos 3-lobed gk",
	TRACK_N,
	track_name,
	track_default,
	track_s,track_e,
	CHECK_N,
	check_name,
	check_default,
	func_proc,
	func_init,
	func_exit,
	NULL,
	func_WndProc,
	NULL,NULL,
	NULL,
	0,
	"Ver 0.5.3 by ܂쏊",
	NULL,
	NULL,
};


/*------------------------------------------------------------------
   export function
 ------------------------------------------------------------------*/
FILTER_DLL __declspec(dllexport) * __stdcall GetFilterTable( void )
{
	return &filter;
}

/*------------------------------------------------------------------
   filter function
 ------------------------------------------------------------------*/
BOOL func_proc(FILTER *fp,FILTER_PROC_INFO *fpip)
{
	PIXEL_YC *swap;

	if(is_setup_required(fp, fpip)){
		__asm {emms};
		setup_table(fp, fpip);
	}

	if(fpip->w != fp->track[0]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_horz_nosimd, fp, fpip);
		}else{
			horizontal_filter(fpip->ycp_edit, fpip->ycp_temp, &x_prm);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->w = fp->track[0];
	}

	if(fpip->h != fp->track[1]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_vert_nosimd, fp, fpip);
		}else{
			vertical_filter(fpip->ycp_edit, fpip->ycp_temp, &y_prm);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->h = fp->track[1];
	}

	return TRUE;
}

/*------------------------------------------------------------------
   filter function sse2 int32 version
 ------------------------------------------------------------------*/
BOOL func_proc_sse2_int(FILTER *fp,FILTER_PROC_INFO *fpip)
{
	PIXEL_YC *swap;

	if(is_setup_required(fp, fpip)){
		__asm {emms};
		setup_sse2_int_table(fp, fpip);
	}

	if(fpip->w != fp->track[0]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_horz_sse2_int, fp, fpip);
		}else{
			horizontal_filter_sse2_int(fpip->ycp_edit, fpip->ycp_temp, x_sse2_int_prm.index, x_sse2_int_prm.alignment_weight, x_sse2_int_prm.x_loop_count, x_sse2_int_prm.y_loop_count, x_sse2_int_prm.tap, x_sse2_int_prm.step, x_sse2_int_prm.src_length_minus_one, x_sse2_int_prm.center_loop_count);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->w = fp->track[0];
	}

	if(fpip->h != fp->track[1]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_vert_sse2_int, fp, fpip);
		}else{
			vertical_filter_sse2_int(fpip->ycp_edit, fpip->ycp_temp, y_sse2_int_prm.index, y_sse2_int_prm.alignment_weight, y_sse2_int_prm.x_loop_count, y_sse2_int_prm.y_loop_count, y_sse2_int_prm.tap, y_sse2_int_prm.step, y_sse2_int_prm.src_length_minus_one, y_sse2_int_prm.center_loop_count);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->h = fp->track[1];
	}

	return TRUE;
}

/*------------------------------------------------------------------
   filter function sse float version
 ------------------------------------------------------------------*/
BOOL func_proc_sse_float(FILTER *fp,FILTER_PROC_INFO *fpip)
{
	PIXEL_YC *swap;

	if(is_setup_required(fp, fpip)){
		__asm {emms};
		setup_sse_table(fp, fpip);
	}

	if(fpip->w != fp->track[0]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_horz_sse_float, fp, fpip);
		}else{
			horizontal_filter_sse_float(fpip->ycp_edit, fpip->ycp_temp, x_sse_prm.index, x_sse_prm.alignment_weight, x_sse_prm.x_loop_count, x_sse_prm.y_loop_count, x_sse_prm.tap, x_sse_prm.step, x_sse_prm.src_length_minus_one, x_sse_prm.center_loop_count);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->w = fp->track[0];
	}

	if(fpip->h != fp->track[1]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_vert_sse_float, fp, fpip);
		}else{
			vertical_filter_sse_float(fpip->ycp_edit, fpip->ycp_temp, y_sse_prm.index, y_sse_prm.alignment_weight, y_sse_prm.x_loop_count, y_sse_prm.y_loop_count, y_sse_prm.tap, y_sse_prm.step, y_sse_prm.src_length_minus_one, y_sse_prm.center_loop_count);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->h = fp->track[1];
	}

	return TRUE;
}

/*------------------------------------------------------------------
   filter function sse2 double version
 ------------------------------------------------------------------*/
BOOL func_proc_sse2_double(FILTER *fp,FILTER_PROC_INFO *fpip)
{
	PIXEL_YC *swap;

	if(is_setup_required(fp, fpip)){
		__asm {emms};
		setup_sse2_double_table(fp, fpip);
	}

	if(fpip->w != fp->track[0]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_horz_sse2_double, fp, fpip);
		}else{
			horizontal_filter_sse2_double(fpip->ycp_edit, fpip->ycp_temp, x_sse2_double_prm.index, x_sse2_double_prm.alignment_weight, x_sse2_double_prm.x_loop_count, x_sse2_double_prm.y_loop_count, x_sse2_double_prm.tap, x_sse2_double_prm.step, x_sse2_double_prm.src_length_minus_one, x_sse2_double_prm.center_loop_count);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->w = fp->track[0];
	}

	if(fpip->h != fp->track[1]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_vert_sse2_double, fp, fpip);
		}else{
			vertical_filter_sse2_double(fpip->ycp_edit, fpip->ycp_temp, y_sse2_double_prm.index, y_sse2_double_prm.alignment_weight, y_sse2_double_prm.x_loop_count, y_sse2_double_prm.y_loop_count, y_sse2_double_prm.tap, y_sse2_double_prm.step, y_sse2_double_prm.src_length_minus_one, y_sse2_double_prm.center_loop_count);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->h = fp->track[1];
	}

	return TRUE;
}

/*------------------------------------------------------------------
   filter function mmx int version
 ------------------------------------------------------------------*/
BOOL func_proc_mmx_int(FILTER *fp,FILTER_PROC_INFO *fpip)
{
	PIXEL_YC *swap;

	if(is_setup_required(fp, fpip)){
		__asm {emms};
		setup_mmx_int_table(fp, fpip);
	}

	if(fpip->w != fp->track[0]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_horz_mmx_int, fp, fpip);
		}else{
			horizontal_filter_mmx_int(fpip->ycp_edit, fpip->ycp_temp, x_mmx_int_prm.index, x_mmx_int_prm.alignment_weight, x_mmx_int_prm.x_loop_count, x_mmx_int_prm.y_loop_count, x_mmx_int_prm.tap, x_mmx_int_prm.step, x_mmx_int_prm.src_length_minus_one, x_mmx_int_prm.center_loop_count);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->w = fp->track[0];
	}

	if(fpip->h != fp->track[1]){
		if(fp->exfunc->exec_multi_thread_func != NULL){
			fp->exfunc->exec_multi_thread_func(proc_vert_mmx_int, fp, fpip);
		}else{
			vertical_filter_mmx_int(fpip->ycp_edit, fpip->ycp_temp, y_mmx_int_prm.index, y_mmx_int_prm.alignment_weight, y_mmx_int_prm.x_loop_count, y_mmx_int_prm.y_loop_count, y_mmx_int_prm.tap, y_mmx_int_prm.step, y_mmx_int_prm.src_length_minus_one, y_mmx_int_prm.center_loop_count);
		}
		swap = fpip->ycp_edit;
		fpip->ycp_edit = fpip->ycp_temp;
		fpip->ycp_temp = swap;
		fpip->h = fp->track[1];
	}

	return TRUE;
}

/*------------------------------------------------------------------
   initialize function
 ------------------------------------------------------------------*/
BOOL func_init(FILTER *fp)
{
	
	int i;
	
	memset(&x_prm, 0, sizeof(PARAMETER));
	memset(&y_prm, 0, sizeof(PARAMETER));
	memset(&x_sse_prm, 0, sizeof(SSE_PARAMETER));
	memset(&y_sse_prm, 0, sizeof(SSE_PARAMETER));
	memset(&x_sse2_int_prm, 0, sizeof(SSE2_INT_PARAMETER));
	memset(&y_sse2_int_prm, 0, sizeof(SSE2_INT_PARAMETER));
	memset(&x_sse2_double_prm, 0, sizeof(SSE2_DOUBLE_PARAMETER));
	memset(&y_sse2_double_prm, 0, sizeof(SSE2_DOUBLE_PARAMETER));
	memset(&x_mmx_int_prm, 0, sizeof(MMX_INT_PARAMETER));
	memset(&y_mmx_int_prm, 0, sizeof(MMX_INT_PARAMETER));

	src_width = 0;
	src_height = 0;
	rst_width = 0;
	rst_height = 0;

	for(i=0;i<4;i++){
		sse_half[i] = (float)0.5;
	}

	for(i=0;i<2;i++){
		sse2_half[i] = 0.5;
	}
	
	UNREFERENCED_PARAMETER(fp);
	
	return TRUE;
}

/*------------------------------------------------------------------
   finalize function
 ------------------------------------------------------------------*/
BOOL func_exit(FILTER *fp)
{
	clear_parameter(&x_prm);
	clear_parameter(&y_prm);
	clear_sse_parameter(&x_sse_prm);
	clear_sse_parameter(&y_sse_prm);
	clear_sse2_int_parameter(&x_sse2_int_prm);
	clear_sse2_int_parameter(&y_sse2_int_prm);
	clear_sse2_double_parameter(&x_sse2_double_prm);
	clear_sse2_double_parameter(&y_sse2_double_prm);
	clear_mmx_int_parameter(&x_mmx_int_prm);
	clear_mmx_int_parameter(&y_mmx_int_prm);

	UNREFERENCED_PARAMETER(fp);
	
	return TRUE;
}

/*-----------------------------------------------------------------*/
static void proc_horz_nosimd(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;
	PARAMETER prm;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = x_prm.height / thread_num;
	mod = x_prm.height % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = fpip->ycp_edit + (head * fpip->max_w);
	dst = fpip->ycp_temp + (head * fpip->max_w);

	memcpy(&prm, &x_prm, sizeof(PARAMETER));
	prm.height = tail - head;

	if(prm.height > 0){
		horizontal_filter(src, dst, &prm);
	}
}

/*-----------------------------------------------------------------*/
static void proc_vert_nosimd(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;
	PARAMETER prm;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = y_prm.width / thread_num;
	mod = y_prm.width % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = fpip->ycp_edit + head;
	dst = fpip->ycp_temp + head;

	memcpy(&prm, &y_prm, sizeof(PARAMETER));
	prm.width = tail - head;

	if(prm.width > 0){
		vertical_filter(src, dst, &prm);
	}
}

/*-----------------------------------------------------------------*/
static void proc_horz_sse2_int(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = x_sse2_int_prm.y_loop_count / thread_num;
	mod = x_sse2_int_prm.y_loop_count % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = fpip->ycp_edit + (head * fpip->max_w);
	dst = fpip->ycp_temp + (head * fpip->max_w);

	if((tail-head) > 0){
		horizontal_filter_sse2_int(src, dst, x_sse2_int_prm.index, x_sse2_int_prm.alignment_weight, x_sse2_int_prm.x_loop_count, tail-head, x_sse2_int_prm.tap, x_sse2_int_prm.step, x_sse2_int_prm.src_length_minus_one, x_sse2_int_prm.center_loop_count);
	}
}

/*-----------------------------------------------------------------*/
static void proc_vert_sse2_int(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = y_sse2_int_prm.x_loop_count / thread_num;
	mod = y_sse2_int_prm.x_loop_count % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = (PIXEL_YC *)(((unsigned char *)fpip->ycp_edit) + (head * 16));
	dst = (PIXEL_YC *)(((unsigned char *)fpip->ycp_temp) + (head * 16));

	if((tail-head) > 0){
		vertical_filter_sse2_int(src, dst, y_sse2_int_prm.index, y_sse2_int_prm.alignment_weight, tail-head, y_sse2_int_prm.y_loop_count, y_sse2_int_prm.tap, y_sse2_int_prm.step, y_sse2_int_prm.src_length_minus_one, y_sse2_int_prm.center_loop_count);
	}
}

/*-----------------------------------------------------------------*/
static void proc_horz_sse_float(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = x_sse_prm.y_loop_count / thread_num;
	mod = x_sse_prm.y_loop_count % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = fpip->ycp_edit + (head * fpip->max_w);
	dst = fpip->ycp_temp + (head * fpip->max_w);

	if((tail-head) > 0){
		horizontal_filter_sse_float(src, dst, x_sse_prm.index, x_sse_prm.alignment_weight, x_sse_prm.x_loop_count, tail-head, x_sse_prm.tap, x_sse_prm.step, x_sse_prm.src_length_minus_one, x_sse_prm.center_loop_count);
	}
}

/*-----------------------------------------------------------------*/
static void proc_vert_sse_float(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = y_sse_prm.x_loop_count / thread_num;
	mod = y_sse_prm.x_loop_count % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = (PIXEL_YC *)(((unsigned char *)fpip->ycp_edit) + (head * 16));
	dst = (PIXEL_YC *)(((unsigned char *)fpip->ycp_temp) + (head * 16));

	if((tail-head) > 0){
		vertical_filter_sse_float(src, dst, y_sse_prm.index, y_sse_prm.alignment_weight, tail-head, y_sse_prm.y_loop_count, y_sse_prm.tap, y_sse_prm.step, y_sse_prm.src_length_minus_one, y_sse_prm.center_loop_count);
	}
}

/*-----------------------------------------------------------------*/
static void proc_horz_sse2_double(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = x_sse2_double_prm.y_loop_count / thread_num;
	mod = x_sse2_double_prm.y_loop_count % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = fpip->ycp_edit + (head * fpip->max_w);
	dst = fpip->ycp_temp + (head * fpip->max_w);

	if((tail-head) > 0){
		horizontal_filter_sse2_double(src, dst, x_sse2_double_prm.index, x_sse2_double_prm.alignment_weight, x_sse2_double_prm.x_loop_count, tail-head, x_sse2_double_prm.tap, x_sse2_double_prm.step, x_sse2_double_prm.src_length_minus_one, x_sse2_double_prm.center_loop_count);
	}
}

/*-----------------------------------------------------------------*/
static void proc_vert_sse2_double(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = y_sse2_double_prm.x_loop_count / thread_num;
	mod = y_sse2_double_prm.x_loop_count % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = (PIXEL_YC *)(((unsigned char *)fpip->ycp_edit) + (head * 8));
	dst = (PIXEL_YC *)(((unsigned char *)fpip->ycp_temp) + (head * 8));

	if((tail-head) > 0){
		vertical_filter_sse2_double(src, dst, y_sse2_double_prm.index, y_sse2_double_prm.alignment_weight, tail-head, y_sse2_double_prm.y_loop_count, y_sse2_double_prm.tap, y_sse2_double_prm.step, y_sse2_double_prm.src_length_minus_one, y_sse2_double_prm.center_loop_count);
	}
}

/*-----------------------------------------------------------------*/
static void proc_horz_mmx_int(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = x_mmx_int_prm.y_loop_count / thread_num;
	mod = x_mmx_int_prm.y_loop_count % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = fpip->ycp_edit + (head * fpip->max_w);
	dst = fpip->ycp_temp + (head * fpip->max_w);

	if((tail-head) > 0){
		horizontal_filter_mmx_int(src, dst, x_mmx_int_prm.index, x_mmx_int_prm.alignment_weight, x_mmx_int_prm.x_loop_count, tail-head, x_mmx_int_prm.tap, x_mmx_int_prm.step, x_mmx_int_prm.src_length_minus_one, x_mmx_int_prm.center_loop_count);
	}
}

/*-----------------------------------------------------------------*/
static void proc_vert_mmx_int(int thread_id, int thread_num, void *param1, void *param2)
{
	int head, tail;
	int unit, mod;
	
	FILTER *fp;
	FILTER_PROC_INFO *fpip;

	PIXEL_YC *src;
	PIXEL_YC *dst;

	fp = (FILTER *)param1;
	fpip = (FILTER_PROC_INFO *)param2;

	unit = y_mmx_int_prm.x_loop_count / thread_num;
	mod = y_mmx_int_prm.x_loop_count % thread_num;
	if(thread_id < mod){
		head = (unit+1)*thread_id;
		tail = head + unit + 1;
	}else{
		head = ((unit+1) * mod) + (unit * (thread_id-mod));
		tail = head + unit;
	}

	src = (PIXEL_YC *)(((unsigned char *)fpip->ycp_edit) + (head * 8));
	dst = (PIXEL_YC *)(((unsigned char *)fpip->ycp_temp) + (head * 8));

	if((tail-head) > 0){
		vertical_filter_mmx_int(src, dst, y_mmx_int_prm.index, y_mmx_int_prm.alignment_weight, tail-head, y_mmx_int_prm.y_loop_count, y_mmx_int_prm.tap, y_mmx_int_prm.step, y_mmx_int_prm.src_length_minus_one, y_mmx_int_prm.center_loop_count);
	}
}

/*-----------------------------------------------------------------*/
static int is_setup_required(FILTER *fp, FILTER_PROC_INFO *fpip)
{
	if(fp->track[0] != rst_width){
		return 1;
	}

	if(fp->track[1] != rst_height){
		return 1;
	}

	if(fpip->w != src_width){
		return 1;
	}

	if(fpip->h != src_height){
		return 1;
	}

	return 0;
}

/*-----------------------------------------------------------------*/
static void setup_table(FILTER *fp, FILTER_PROC_INFO *fpip)
{
	rst_width = fp->track[0];
	rst_height = fp->track[1];
	src_width = fpip->w;
	src_height = fpip->h;

	x_prm.width = rst_width;
	x_prm.height = src_height;

	y_prm.width = rst_width;
	y_prm.height = rst_height;

	x_prm.step = fpip->max_w;
	y_prm.step = fpip->max_w;

	if(rst_width > src_width){
		setup_interpolation_parameter(src_width, rst_width, &x_prm);
	}else if(rst_width < src_width){
		setup_decimation_parameter(src_width, rst_width, &x_prm);
	}

	if(rst_height > src_height){
		setup_interpolation_parameter(src_height, rst_height, &y_prm);
	}else if(rst_height < src_height){
		setup_decimation_parameter(src_height, rst_height, &y_prm);
	}
}


/*-----------------------------------------------------------------*/
static void setup_sse_table(FILTER *fp, FILTER_PROC_INFO *fpip)
{
	rst_width = fp->track[0];
	rst_height = fp->track[1];
	src_width = fpip->w;
	src_height = fpip->h;

	x_sse_prm.x_loop_count = rst_width;
	x_sse_prm.y_loop_count = src_height;
	x_sse_prm.step = fpip->max_w * 6;

	y_sse_prm.x_loop_count = (rst_width * 3 + 7) / 8;
	y_sse_prm.y_loop_count = rst_height;
	y_sse_prm.step = fpip->max_w * 6;

	if(rst_width > src_width){
		setup_sse_interpolation_parameter(src_width, rst_width, &x_sse_prm);
	}else if(rst_width < src_width){
		setup_sse_decimation_parameter(src_width, rst_width, &x_sse_prm);
	}

	if(rst_height > src_height){
		setup_sse_interpolation_parameter(src_height, rst_height, &y_sse_prm);
	}else if(rst_height < src_height){
		setup_sse_decimation_parameter(src_height, rst_height, &y_sse_prm);
	}
}


/*-----------------------------------------------------------------*/
static void setup_sse2_int_table(FILTER *fp, FILTER_PROC_INFO *fpip)
{
	rst_width = fp->track[0];
	rst_height = fp->track[1];
	src_width = fpip->w;
	src_height = fpip->h;

	x_sse2_int_prm.x_loop_count = rst_width;
	x_sse2_int_prm.y_loop_count = src_height;
	x_sse2_int_prm.step = fpip->max_w * 6;

	y_sse2_int_prm.x_loop_count = (rst_width * 3 + 7) / 8;
	y_sse2_int_prm.y_loop_count = rst_height;
	y_sse2_int_prm.step = fpip->max_w * 6;

	if(rst_width > src_width){
		setup_sse2_int_interpolation_parameter(src_width, rst_width, &x_sse2_int_prm);
	}else if(rst_width < src_width){
		setup_sse2_int_decimation_parameter(src_width, rst_width, &x_sse2_int_prm);
	}

	if(rst_height > src_height){
		setup_sse2_int_interpolation_parameter(src_height, rst_height, &y_sse2_int_prm);
	}else if(rst_height < src_height){
		setup_sse2_int_decimation_parameter(src_height, rst_height, &y_sse2_int_prm);
	}
}

/*-----------------------------------------------------------------*/
static void setup_sse2_double_table(FILTER *fp, FILTER_PROC_INFO *fpip)
{
	rst_width = fp->track[0];
	rst_height = fp->track[1];
	src_width = fpip->w;
	src_height = fpip->h;

	x_sse2_double_prm.x_loop_count = rst_width;
	x_sse2_double_prm.y_loop_count = src_height;
	x_sse2_double_prm.step = fpip->max_w * 6;

	y_sse2_double_prm.x_loop_count = (rst_width * 3 + 3) / 4;
	y_sse2_double_prm.y_loop_count = rst_height;
	y_sse2_double_prm.step = fpip->max_w * 6;

	if(rst_width > src_width){
		setup_sse2_double_interpolation_parameter(src_width, rst_width, &x_sse2_double_prm);
	}else if(rst_width < src_width){
		setup_sse2_double_decimation_parameter(src_width, rst_width, &x_sse2_double_prm);
	}

	if(rst_height > src_height){
		setup_sse2_double_interpolation_parameter(src_height, rst_height, &y_sse2_double_prm);
	}else if(rst_height < src_height){
		setup_sse2_double_decimation_parameter(src_height, rst_height, &y_sse2_double_prm);
	}
}

/*-----------------------------------------------------------------*/
static void setup_mmx_int_table(FILTER *fp, FILTER_PROC_INFO *fpip)
{
	rst_width = fp->track[0];
	rst_height = fp->track[1];
	src_width = fpip->w;
	src_height = fpip->h;

	x_mmx_int_prm.x_loop_count = rst_width;
	x_mmx_int_prm.y_loop_count = src_height;
	x_mmx_int_prm.step = fpip->max_w * 6;

	y_mmx_int_prm.x_loop_count = (rst_width * 3 + 3) / 4;
	y_mmx_int_prm.y_loop_count = rst_height;
	y_mmx_int_prm.step = fpip->max_w * 6;

	if(rst_width > src_width){
		setup_mmx_int_interpolation_parameter(src_width, rst_width, &x_mmx_int_prm);
	}else if(rst_width < src_width){
		setup_mmx_int_decimation_parameter(src_width, rst_width, &x_mmx_int_prm);
	}

	if(rst_height > src_height){
		setup_mmx_int_interpolation_parameter(src_height, rst_height, &y_mmx_int_prm);
	}else if(rst_height < src_height){
		setup_mmx_int_decimation_parameter(src_height, rst_height, &y_mmx_int_prm);
	}
}

/*-----------------------------------------------------------------*/
static void setup_interpolation_parameter(int source_length, int result_length, PARAMETER *out)
{
	int i,j,n;
	double *work;
	double  sum;
	double  pos;

	clear_parameter(out);
	
	out->length = result_length;
	out->weight = (int **)malloc(sizeof(int *)*out->length);
	out->index = (int **)malloc(sizeof(int *)*out->length);
	out->tap = 6;

	for(i=0;i<result_length;i++){
		out->weight[i] = (int *)malloc(sizeof(int)*out->tap);
		out->index[i] = (int *)malloc(sizeof(int)*out->tap);
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	for(i=0;i<result_length;i++){
		pos = (i+0.5)*source_length;
		pos /= result_length;
		n = floor(pos-2.5);
		pos = (n+0.5-pos);
		sum = 0;
		for(j=0;j<out->tap;j++){
			if(n < 0){
				out->index[i][j] = 0;
			}else if(n >= source_length){
				out->index[i][j] = source_length-1;
			}else{
				out->index[i][j] = n;
			}
			work[j] = lanczos3_weight(pos);
			sum += work[j];
			pos += 1;
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->weight[i][j] = (int)((work[j] / sum) * (1<<16));
		}
	}

	free(work);
}

/*-----------------------------------------------------------------*/
static void setup_sse_interpolation_parameter(int source_length, int result_length, SSE_PARAMETER *out)
{
	int i,j,n;
	int edge;
	unsigned int pointer;
	double *work;
	double  sum;
	double  pos;

	clear_sse_parameter(out);
	
	out->length = result_length;
	out->src_length_minus_one = source_length - 1;
	out->weight = (float **)malloc(sizeof(float *)*out->length);
	out->alignment_weight = (float **)malloc(sizeof(float *)*out->length);
	out->index = (int *)malloc(sizeof(int)*out->length);
	out->tap = 6;

	for(i=0;i<result_length;i++){
		out->weight[i] = (float *)malloc(sizeof(float)*4*(out->tap+1));
		pointer = (unsigned int) out->weight[i];
		pointer += (16 - (pointer & 0x0f));
		out->alignment_weight[i] = (float *)pointer;
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	out->center_loop_count = 0;

	for(i=0;i<result_length;i++){
		pos = (i+0.5)*source_length;
		pos /= result_length;
		n = floor(pos-2.5);
		pos = (n+0.5-pos);
		sum = 0;
		out->index[i] = n;
		edge = 0;
		for(j=0;j<out->tap;j++){
			if(n < 0){
				edge = 1;
			}else if(n >= source_length){
				edge = 1;
			}
			work[j] = lanczos3_weight(pos);
			sum += work[j];
			pos += 1;
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->alignment_weight[i][j*4] = (float) (work[j] / sum);
			out->alignment_weight[i][j*4+1] = out->alignment_weight[i][j*4];
			out->alignment_weight[i][j*4+2] = out->alignment_weight[i][j*4];
			out->alignment_weight[i][j*4+3] = out->alignment_weight[i][j*4];
		}

		if(edge == 0){
			out->center_loop_count += 1;
		}
	}

	free(work);

}

/*-----------------------------------------------------------------*/
static void setup_sse2_int_interpolation_parameter(int source_length, int result_length, SSE2_INT_PARAMETER *out)
{
	int i,j,n;
	int edge;
	unsigned int pointer;
	double *work;
	double  sum;
	double  pos;

	clear_sse2_int_parameter(out);
	
	out->length = result_length;
	out->src_length_minus_one = source_length - 1;
	out->weight = (int **)malloc(sizeof(int *)*out->length);
	out->alignment_weight = (int **)malloc(sizeof(int *)*out->length);
	out->index = (int *)malloc(sizeof(int)*out->length);
	out->tap = 6;

	for(i=0;i<result_length;i++){
		out->weight[i] = (int *)malloc(sizeof(int)*4*(out->tap+1));
		pointer = (unsigned int) out->weight[i];
		pointer += (16 - (pointer & 0x0f));
		out->alignment_weight[i] = (int *)pointer;
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	out->center_loop_count = 0;

	for(i=0;i<result_length;i++){
		pos = (i+0.5)*source_length;
		pos /= result_length;
		n = floor(pos-2.5);
		pos = (n+0.5-pos);
		sum = 0;
		out->index[i] = n;
		edge = 0;
		for(j=0;j<out->tap;j++){
			if(n < 0){
				edge = 1;
			}else if(n >= source_length){
				edge = 1;
			}
			work[j] = lanczos3_weight(pos);
			sum += work[j];
			pos += 1;
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->alignment_weight[i][j*4] = (int) ((work[j] / sum) * (1<<14));
			out->alignment_weight[i][j*4+1] = out->alignment_weight[i][j*4];
			out->alignment_weight[i][j*4+2] = out->alignment_weight[i][j*4];
			out->alignment_weight[i][j*4+3] = out->alignment_weight[i][j*4];
		}

		if(edge == 0){
			out->center_loop_count += 1;
		}
	}

	free(work);

}

/*-----------------------------------------------------------------*/
static void setup_sse2_double_interpolation_parameter(int source_length, int result_length, SSE2_DOUBLE_PARAMETER *out)
{
	int i,j,n;
	int edge;
	unsigned int pointer;
	double *work;
	double  sum;
	double  pos;

	clear_sse2_double_parameter(out);
	
	out->length = result_length;
	out->src_length_minus_one = source_length - 1;
	out->weight = (double **)malloc(sizeof(double *)*out->length);
	out->alignment_weight = (double **)malloc(sizeof(double *)*out->length);
	out->index = (int *)malloc(sizeof(int)*out->length);
	out->tap = 6;

	for(i=0;i<result_length;i++){
		out->weight[i] = (double *)malloc(sizeof(double)*2*(out->tap+1));
		pointer = (unsigned int) out->weight[i];
		pointer += (16 - (pointer & 0x0f));
		out->alignment_weight[i] = (double *)pointer;
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	out->center_loop_count = 0;

	for(i=0;i<result_length;i++){
		pos = (i+0.5)*source_length;
		pos /= result_length;
		n = floor(pos-2.5);
		pos = (n+0.5-pos);
		sum = 0;
		out->index[i] = n;
		edge = 0;
		for(j=0;j<out->tap;j++){
			if(n < 0){
				edge = 1;
			}else if(n >= source_length){
				edge = 1;
			}
			work[j] = lanczos3_weight(pos);
			sum += work[j];
			pos += 1;
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->alignment_weight[i][j*2] = (work[j] / sum);
			out->alignment_weight[i][j*2+1] = out->alignment_weight[i][j*2];
		}

		if(edge == 0){
			out->center_loop_count += 1;
		}
	}

	free(work);

}

/*-----------------------------------------------------------------*/
static void setup_mmx_int_interpolation_parameter(int source_length, int result_length, MMX_INT_PARAMETER *out)
{
	int i,j,n;
	int edge;
	unsigned int pointer;
	double *work;
	double  sum;
	double  pos;

	clear_mmx_int_parameter(out);
	
	out->length = result_length;
	out->src_length_minus_one = source_length - 1;
	out->weight = (int **)malloc(sizeof(int *)*out->length);
	out->alignment_weight = (int **)malloc(sizeof(int *)*out->length);
	out->index = (int *)malloc(sizeof(int)*out->length);
	out->tap = 6;

	for(i=0;i<result_length;i++){
		out->weight[i] = (int *)malloc(sizeof(int)*2*(out->tap+1));
		pointer = (unsigned int) out->weight[i];
		pointer += (8 - (pointer & 0x07));
		out->alignment_weight[i] = (int *)pointer;
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	out->center_loop_count = 0;

	for(i=0;i<result_length;i++){
		pos = (i+0.5)*source_length;
		pos /= result_length;
		n = floor(pos-2.5);
		pos = (n+0.5-pos);
		sum = 0;
		out->index[i] = n;
		edge = 0;
		for(j=0;j<out->tap;j++){
			if(n < 0){
				edge = 1;
			}else if(n >= source_length){
				edge = 1;
			}
			work[j] = lanczos3_weight(pos);
			sum += work[j];
			pos += 1;
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->alignment_weight[i][j*2] = (int)((work[j] / sum) * (1 << 14));
			out->alignment_weight[i][j*2+1] = out->alignment_weight[i][j*2];
		}

		if(edge == 0){
			out->center_loop_count += 1;
		}
	}

	free(work);

}

/*-----------------------------------------------------------------*/
static void setup_decimation_parameter(int source_length, int result_length, PARAMETER *out)
{
	int i,j,n;
	double *work;
	double  sum;
	double  pos, phase;

	clear_parameter(out);

	out->length = result_length;
	out->weight = (int **)malloc(sizeof(int *)*out->length);
	out->index = (int **)malloc(sizeof(int *)*out->length);
	
	out->tap = (6*(source_length)+(result_length-1)) / result_length;

	for(i=0;i<result_length;i++){
		out->weight[i] = (int *)malloc(sizeof(int)*out->tap);
		out->index[i] = (int *)malloc(sizeof(int)*out->tap);
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	for(i=0;i<result_length;i++){
		pos = (i-3+0.5)*source_length/result_length + 0.5;
		n = floor(pos);
		sum = 0;
		for(j=0;j<out->tap;j++){
			phase = (n+0.5)*result_length;
			phase /= source_length;
			phase -= (i+0.5);
			if(n < 0){
				out->index[i][j] = 0;
			}else if(n >= source_length){
				out->index[i][j] = source_length-1;
			}else{
				out->index[i][j] = n;
			}
			work[j] = lanczos3_weight(phase);
			sum += work[j];
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->weight[i][j] = (int)((work[j] / sum) * (1<<16));
		}
	}

	free(work);
}

/*-----------------------------------------------------------------*/
static void setup_sse_decimation_parameter(int source_length, int result_length, SSE_PARAMETER *out)
{
	int i,j,n;
	int edge;
	unsigned int pointer;
	double *work;
	double  sum;
	double  pos, phase;

	clear_sse_parameter(out);

	out->length = result_length;
	out->src_length_minus_one = source_length - 1;
	out->weight = (float **)malloc(sizeof(float *)*out->length);
	out->alignment_weight = (float **)malloc(sizeof(float *)*out->length);
	out->index = (int *)malloc(sizeof(int)*out->length);
	
	out->tap = (6*(source_length)+(result_length-1)) / result_length;

	for(i=0;i<result_length;i++){
		out->weight[i] = (float *)malloc(sizeof(float)*4*(out->tap+1));
		pointer = (unsigned int)out->weight[i];
		pointer += (16 - (pointer & 0x0f));
		out->alignment_weight[i] = (float *)pointer;
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	out->center_loop_count = 0;

	for(i=0;i<result_length;i++){
		pos = (i-3+0.5)*source_length/result_length + 0.5;
		n = floor(pos);
		sum = 0;
		edge = 0;
		out->index[i] = n;
		for(j=0;j<out->tap;j++){
			phase = (n+0.5)*result_length;
			phase /= source_length;
			phase -= (i+0.5);
			if(n < 0){
				edge = 1;
			}else if(n >= source_length){
				edge = 1;
			}
			work[j] = lanczos3_weight(phase);
			sum += work[j];
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->alignment_weight[i][j*4] = (float) (work[j] / sum);
			out->alignment_weight[i][j*4+1] = out->alignment_weight[i][j*4];
			out->alignment_weight[i][j*4+2] = out->alignment_weight[i][j*4];
			out->alignment_weight[i][j*4+3] = out->alignment_weight[i][j*4];
		}
		
		if(edge == 0){
			out->center_loop_count += 1;
		}
	}

	free(work);
}

/*-----------------------------------------------------------------*/
static void setup_sse2_int_decimation_parameter(int source_length, int result_length, SSE2_INT_PARAMETER *out)
{
	int i,j,n;
	int edge;
	unsigned int pointer;
	double *work;
	double  sum;
	double  pos, phase;

	clear_sse2_int_parameter(out);

	out->length = result_length;
	out->src_length_minus_one = source_length - 1;
	out->weight = (int **)malloc(sizeof(int *)*out->length);
	out->alignment_weight = (int **)malloc(sizeof(int *)*out->length);
	out->index = (int *)malloc(sizeof(int)*out->length);
	
	out->tap = (6*(source_length)+(result_length-1)) / result_length;

	for(i=0;i<result_length;i++){
		out->weight[i] = (int *)malloc(sizeof(int)*4*(out->tap+1));
		pointer = (unsigned int) out->weight[i];
		pointer += (16 - (pointer & 0x0f));
		out->alignment_weight[i] = (int *)pointer;
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	out->center_loop_count = 0;

	for(i=0;i<result_length;i++){
		pos = (i-3+0.5)*source_length/result_length + 0.5;
		n = floor(pos);
		sum = 0;
		edge = 0;
		out->index[i] = n;
		for(j=0;j<out->tap;j++){
			phase = (n+0.5)*result_length;
			phase /= source_length;
			phase -= (i+0.5);
			if(n < 0){
				edge = 1;
			}else if(n >= source_length){
				edge = 1;
			}
			work[j] = lanczos3_weight(phase);
			sum += work[j];
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->alignment_weight[i][j*4] = (int) ((work[j] / sum) * (1<<14));
			out->alignment_weight[i][j*4+1] = out->alignment_weight[i][j*4];
			out->alignment_weight[i][j*4+2] = out->alignment_weight[i][j*4];
			out->alignment_weight[i][j*4+3] = out->alignment_weight[i][j*4];
		}
		
		if(edge == 0){
			out->center_loop_count += 1;
		}
	}

	free(work);
}

/*-----------------------------------------------------------------*/
static void setup_sse2_double_decimation_parameter(int source_length, int result_length, SSE2_DOUBLE_PARAMETER *out)
{
	int i,j,n;
	int edge;
	unsigned int pointer;
	double *work;
	double  sum;
	double  pos, phase;

	clear_sse2_double_parameter(out);

	out->length = result_length;
	out->src_length_minus_one = source_length - 1;
	out->weight = (double **)malloc(sizeof(double *)*out->length);
	out->alignment_weight = (double **)malloc(sizeof(double *)*out->length);
	out->index = (int *)malloc(sizeof(int)*out->length);
	
	out->tap = (6*(source_length)+(result_length-1)) / result_length;

	for(i=0;i<result_length;i++){
		out->weight[i] = (double *)malloc(sizeof(double)*2*(out->tap+1));
		pointer = (unsigned int) out->weight[i];
		pointer += (16 - (pointer & 0x0f));
		out->alignment_weight[i] = (double *)pointer;
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	out->center_loop_count = 0;

	for(i=0;i<result_length;i++){
		pos = (i-3+0.5)*source_length/result_length + 0.5;
		n = floor(pos);
		sum = 0;
		edge = 0;
		out->index[i] = n;
		for(j=0;j<out->tap;j++){
			phase = (n+0.5)*result_length;
			phase /= source_length;
			phase -= (i+0.5);
			if(n < 0){
				edge = 1;
			}else if(n >= source_length){
				edge = 1;
			}
			work[j] = lanczos3_weight(phase);
			sum += work[j];
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->alignment_weight[i][j*2] = (work[j] / sum);
			out->alignment_weight[i][j*2+1] = out->alignment_weight[i][j*2];
		}
		
		if(edge == 0){
			out->center_loop_count += 1;
		}
	}

	free(work);
}

/*-----------------------------------------------------------------*/
static void setup_mmx_int_decimation_parameter(int source_length, int result_length, MMX_INT_PARAMETER *out)
{
	int i,j,n;
	int edge;
	unsigned int pointer;
	double *work;
	double  sum;
	double  pos, phase;

	clear_mmx_int_parameter(out);

	out->length = result_length;
	out->src_length_minus_one = source_length - 1;
	out->weight = (int **)malloc(sizeof(int *)*out->length);
	out->alignment_weight = (int **)malloc(sizeof(int *)*out->length);
	out->index = (int *)malloc(sizeof(int)*out->length);
	
	out->tap = (6*(source_length)+(result_length-1)) / result_length;

	for(i=0;i<result_length;i++){
		out->weight[i] = (int *)malloc(sizeof(int)*2*(out->tap+1));
		pointer = (unsigned int) out->weight[i];
		pointer += (8 - (pointer & 0x07));
		out->alignment_weight[i] = (int *)pointer;
	}
	work = (double *)malloc(sizeof(double)*out->tap);

	out->center_loop_count = 0;

	for(i=0;i<result_length;i++){
		pos = (i-3+0.5)*source_length/result_length + 0.5;
		n = floor(pos);
		sum = 0;
		edge = 0;
		out->index[i] = n;
		for(j=0;j<out->tap;j++){
			phase = (n+0.5)*result_length;
			phase /= source_length;
			phase -= (i+0.5);
			if(n < 0){
				edge = 1;
			}else if(n >= source_length){
				edge = 1;
			}
			work[j] = lanczos3_weight(phase);
			sum += work[j];
			n += 1;
		}

		for(j=0;j<out->tap;j++){
			out->alignment_weight[i][j*2] = (int) ((work[j] / sum) * (1<<14));
			out->alignment_weight[i][j*2+1] = out->alignment_weight[i][j*2];
		}
		
		if(edge == 0){
			out->center_loop_count += 1;
		}
	}

	free(work);
}

/*-----------------------------------------------------------------*/
static void clear_parameter(PARAMETER *prm)
{
	int i;

	if(prm->tap){
		for(i=0;i<prm->length;i++){
			free(prm->index[i]);
			free(prm->weight[i]);
		}
	}

	if(prm->length){
		free(prm->index);
		free(prm->weight);
	}

	prm->length = 0;
	prm->tap = 0;
	prm->index = NULL;
	prm->weight = NULL;
}

/*-----------------------------------------------------------------*/
static void clear_sse_parameter(SSE_PARAMETER *prm)
{
	int i;

	if(prm->tap){
		for(i=0;i<prm->length;i++){
			free(prm->weight[i]);
		}
	}

	if(prm->length){
		free(prm->index);
		free(prm->weight);
		free(prm->alignment_weight);
	}

	prm->length = 0;
	prm->tap = 0;
	prm->center_loop_count = 0;
	prm->src_length_minus_one = 0;
	prm->index = NULL;
	prm->weight = NULL;
	prm->alignment_weight = NULL;
}

/*-----------------------------------------------------------------*/
static void clear_sse2_int_parameter(SSE2_INT_PARAMETER *prm)
{
	int i;

	if(prm->tap){
		for(i=0;i<prm->length;i++){
			free(prm->weight[i]);
		}
	}

	if(prm->length){
		free(prm->index);
		free(prm->weight);
		free(prm->alignment_weight);
	}

	prm->length = 0;
	prm->tap = 0;
	prm->center_loop_count = 0;
	prm->src_length_minus_one = 0;
	prm->index = NULL;
	prm->weight = NULL;
	prm->alignment_weight = NULL;
}

/*-----------------------------------------------------------------*/
static void clear_sse2_double_parameter(SSE2_DOUBLE_PARAMETER *prm)
{
	int i;

	if(prm->tap){
		for(i=0;i<prm->length;i++){
			free(prm->weight[i]);
		}
	}

	if(prm->length){
		free(prm->index);
		free(prm->weight);
		free(prm->alignment_weight);
	}

	prm->length = 0;
	prm->tap = 0;
	prm->center_loop_count = 0;
	prm->src_length_minus_one = 0;
	prm->index = NULL;
	prm->weight = NULL;
	prm->alignment_weight = NULL;
}

/*-----------------------------------------------------------------*/
static void clear_mmx_int_parameter(MMX_INT_PARAMETER *prm)
{
	int i;

	if(prm->tap){
		for(i=0;i<prm->length;i++){
			free(prm->weight[i]);
		}
	}

	if(prm->length){
		free(prm->index);
		free(prm->weight);
		free(prm->alignment_weight);
	}

	prm->length = 0;
	prm->tap = 0;
	prm->center_loop_count = 0;
	prm->src_length_minus_one = 0;
	prm->index = NULL;
	prm->weight = NULL;
	prm->alignment_weight = NULL;
}

/*-----------------------------------------------------------------*/
static double lanczos3_weight(double phase)
{
	double ret;
	
	if(fabs(phase) < DBL_EPSILON){
		return 1.0;
	}

	if(fabs(phase) >= 3.0){
		return 0.0;
	}

	ret = sin(PI*phase)*sin(PI*phase/3)/(PI*PI*phase*phase/3);

	return ret;
}


/*-----------------------------------------------------------------*/
static void horizontal_filter(PIXEL_YC *in, PIXEL_YC *out, PARAMETER *prm)
{
	int i,j,k;

	int luma, cb, cr;

	for(i=0;i<prm->height;i++){
		for(j=0;j<prm->width;j++){
			luma = 0;
			cb = 0;
			cr = 0;
			for(k=0;k<prm->tap;k++){
				luma += in[prm->index[j][k]+i*prm->step].y * prm->weight[j][k];
				cb += in[prm->index[j][k]+i*prm->step].cb * prm->weight[j][k];
				cr += in[prm->index[j][k]+i*prm->step].cr * prm->weight[j][k];
			}
			out[i*prm->step+j].y = (short)((luma + (1 << 15)) >> 16);
			out[i*prm->step+j].cb = (short)((cb + (1 << 15)) >> 16);
			out[i*prm->step+j].cr = (short)((cr + (1 << 15)) >> 16);
		}
	}
}

/*-----------------------------------------------------------------*/
static void vertical_filter(PIXEL_YC *in, PIXEL_YC *out, PARAMETER *prm)
{
	int i,j,k;

	int luma, cb, cr;

	for(i=0;i<prm->height;i++){
		for(j=0;j<prm->width;j++){
			luma = 0;
			cb = 0;
			cr = 0;
			for(k=0;k<prm->tap;k++){
				luma += in[prm->index[i][k]*prm->step+j].y * prm->weight[i][k];
				cb += in[prm->index[i][k]*prm->step+j].cb * prm->weight[i][k];
				cr += in[prm->index[i][k]*prm->step+j].cr * prm->weight[i][k];
			}
			out[i*prm->step+j].y = (short)((luma + (1 << 15)) >> 16);
			out[i*prm->step+j].cb = (short)((cb + (1 << 15)) >> 16);
			out[i*prm->step+j].cr = (short)((cr + (1 << 15)) >> 16);
		}
	}
}

/*-----------------------------------------------------------------*/
static int is_sse2_enable()
{
	DWORD sse2_check;

	__asm{
		mov eax, 1
		cpuid
		mov sse2_check, edx
	}

	if(sse2_check & 0x04000000){
		_try{
			__asm{
				pxor xmm0, xmm0 ;
			}
		}
		_except(EXCEPTION_EXECUTE_HANDLER){
			if (_exception_code()==STATUS_ILLEGAL_INSTRUCTION){
				return 0;
			}
		}
		return 1;
	}else{
		return 0;
	}
}

/*-----------------------------------------------------------------*/
static int is_sse_enable()
{
	DWORD sse_check;

	__asm{
		mov eax, 1
		cpuid
		mov sse_check, edx
	}

	if(sse_check & 0x02000000){
		_try{
			__asm{
				xorps xmm0, xmm0 ;
			}
		}
		_except(EXCEPTION_EXECUTE_HANDLER){
			if (_exception_code()==STATUS_ILLEGAL_INSTRUCTION){
				return 0;
			}
		}
		return 1;
	}else{
		return 0;
	}
}

/*-----------------------------------------------------------------*/
static int is_mmx_enable()
{
    DWORD mmx_check;


    _asm{
        mov eax,1;
        cpuid;
        mov mmx_check,edx;
    }

    if(mmx_check & 0x00800000){
        return 1;
    }else{
        return 0;
    }
}

/*------------------------------------------------------------------
   window procedure
 ------------------------------------------------------------------*/
BOOL func_WndProc( HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam, void *editp, FILTER *fp )
{
	switch(message){
	case WM_FILTER_INIT:
		init_dialog(hwnd, fp->dll_hinst);
		setup_range(fp, editp);
		load_ini_value(fp);
		set_subclass();
		enable_simd_radio_button();
		return TRUE;
	case WM_COMMAND:
		switch(LOWORD(wparam)){
		case ID_BUTTON_NO_SIMD:
			on_no_simd_button(fp);
			break;
		case ID_BUTTON_SSE2INT:
			on_sse2_int_button(fp);
			break;
		case ID_BUTTON_SSE_FLT:
			on_sse_float_button(fp);
			break;
		case ID_BUTTON_SSE2DBL:
			on_sse2_double_button(fp);
			break;
		case ID_BUTTON_MMX_INT:
			on_mmx_int_button(fp);
			break;
		case ID_BUTTON_640X480:
			on_640x480_button(fp);
			break;
		case ID_BUTTON_352X240:
			on_352x240_button(fp);
			break;
		case ID_BUTTON_320X240:
			on_320x240_button(fp);
			break;
		case ID_BUTTON_SETSIZE:
			on_setsize_button(fp);
			break;
		default:
			return FALSE;
		}
		return fp->exfunc->is_filter_active(fp);
	case WM_FILTER_EXIT:
		remove_subclass();
		save_ini_value(fp);
		DeleteObject(dialog.font);
		break;
	case WM_KEYUP:
	case WM_KEYDOWN:
	case WM_MOUSEWHEEL:
		SendMessage(GetWindow(hwnd, GW_OWNER), message, wparam, lparam);
		break;
	}
	
	UNREFERENCED_PARAMETER(editp);
	UNREFERENCED_PARAMETER(lparam);
	
	return FALSE;
}

/*-----------------------------------------------------------------*/
static void init_dialog(HWND hwnd, HINSTANCE hinst)
{
	dialog.font = CreateFont(12, 0, 0, 0, FW_MEDIUM, FALSE, FALSE, FALSE, SHIFTJIS_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, PROOF_QUALITY, DEFAULT_PITCH | FF_MODERN, "lr oSVbN");
    
	dialog.simd_group = CreateWindow("BUTTON", "SIMD", WS_CHILD|WS_VISIBLE|BS_GROUPBOX, 7, 62, 300, 66, hwnd, NULL, hinst, NULL);
	SendMessage(dialog.simd_group, WM_SETFONT, (WPARAM)dialog.font, 0);

	dialog.rb_no_simd = CreateWindow("BUTTON", " SIMD", WS_CHILD|WS_VISIBLE|WS_GROUP|WS_TABSTOP|BS_AUTORADIOBUTTON, 16, 80, 90, 18, hwnd, (HMENU)ID_BUTTON_NO_SIMD, hinst, NULL);
	SendMessage(dialog.rb_no_simd, WM_SETFONT, (WPARAM)dialog.font, 0);

	dialog.rb_sse2_int = CreateWindow("BUTTON", "SSE2 ", WS_CHILD|WS_VISIBLE|WS_GROUP|WS_TABSTOP|BS_AUTORADIOBUTTON, 110, 80, 90, 18, hwnd, (HMENU)ID_BUTTON_SSE2INT, hinst, NULL);
	SendMessage(dialog.rb_sse2_int, WM_SETFONT, (WPARAM)dialog.font, 0);

	dialog.rb_sse_float = CreateWindow("BUTTON", "SSE ", WS_CHILD|WS_VISIBLE|BS_AUTORADIOBUTTON, 204, 80, 90, 18, hwnd, (HMENU)ID_BUTTON_SSE_FLT, hinst, NULL);
	SendMessage(dialog.rb_sse_float, WM_SETFONT, (WPARAM)dialog.font, 0);

	dialog.rb_sse2_double = CreateWindow("BUTTON", "SSE2 ", WS_CHILD|WS_VISIBLE|BS_AUTORADIOBUTTON, 16, 100, 90, 18, hwnd, (HMENU)ID_BUTTON_SSE2DBL, hinst, NULL);
	SendMessage(dialog.rb_sse2_double, WM_SETFONT, (WPARAM)dialog.font, 0);
	
	dialog.rb_mmx_int = CreateWindow("BUTTON", "MMX ",  WS_CHILD|WS_VISIBLE|BS_AUTORADIOBUTTON, 110, 100, 90, 18, hwnd, (HMENU)ID_BUTTON_MMX_INT, hinst, NULL);
	SendMessage(dialog.rb_mmx_int, WM_SETFONT, (WPARAM)dialog.font, 0);

	dialog.b640x480 = CreateWindow("BUTTON", "640x480", WS_CHILD|WS_VISIBLE|WS_GROUP|WS_TABSTOP|BS_PUSHBUTTON|BS_VCENTER, 7, 134, 52, 18, hwnd, (HMENU)ID_BUTTON_640X480, hinst, NULL);
	SendMessage(dialog.b640x480, WM_SETFONT, (WPARAM)dialog.font, 0);
		
	dialog.b352x240 = CreateWindow("BUTTON", "352x240", WS_CHILD|WS_VISIBLE|BS_PUSHBUTTON|BS_VCENTER, 62, 134, 52, 18, hwnd, (HMENU)ID_BUTTON_352X240, hinst, NULL);
	SendMessage(dialog.b352x240, WM_SETFONT, (WPARAM)dialog.font, 0);
    
	dialog.b320x240 = CreateWindow("BUTTON", "320x240", WS_CHILD|WS_VISIBLE|BS_PUSHBUTTON|BS_VCENTER, 117, 134, 52, 18, hwnd, (HMENU)ID_BUTTON_320X240, hinst, NULL);
	SendMessage(dialog.b320x240, WM_SETFONT, (WPARAM)dialog.font, 0);

	dialog.b_set_size = CreateWindow("BUTTON", "w ->", WS_CHILD|WS_VISIBLE|BS_PUSHBUTTON|BS_VCENTER, 172, 134, 50, 18, hwnd, (HMENU)ID_BUTTON_SETSIZE, hinst, NULL);
	SendMessage(dialog.b_set_size, WM_SETFONT, (WPARAM)dialog.font, 0);

	dialog.e_width = CreateWindowEx(WS_EX_CLIENTEDGE, "EDIT", NULL, WS_CHILD|WS_VISIBLE|ES_RIGHT|ES_NUMBER, 225, 134, 35, 18, hwnd, NULL, hinst, NULL);
	SendMessage(dialog.e_width, WM_SETFONT, (WPARAM)dialog.font, 0);
	SendMessage(dialog.e_width, EM_SETLIMITTEXT, (WPARAM)4, 0);
		
	dialog.e_height = CreateWindowEx(WS_EX_CLIENTEDGE, "EDIT", NULL, WS_CHILD|WS_VISIBLE|ES_RIGHT|ES_NUMBER, 271, 134, 35, 18, hwnd, NULL, hinst, NULL);
	SendMessage(dialog.e_height, WM_SETFONT, (WPARAM)dialog.font, 0);
	SendMessage(dialog.e_height, EM_SETLIMITTEXT, (WPARAM)4, 0);

	dialog.label = CreateWindow("STATIC", "x", WS_CHILD|WS_VISIBLE, 262, 134, 7, 18, hwnd, NULL, hinst, NULL);
	SendMessage(dialog.label, WM_SETFONT, (WPARAM)dialog.font, 0);
}

/*-----------------------------------------------------------------*/
static void set_subclass()
{
	default_button_window_procedure = (WNDPROC)SetWindowLong(dialog.rb_sse2_int, GWL_WNDPROC, (LONG)sse2_button_window_procedure);
	SetWindowLong(dialog.rb_sse_float, GWL_WNDPROC, (LONG)sse_button_window_procedure);
	SetWindowLong(dialog.rb_sse2_double, GWL_WNDPROC, (LONG)sse2_button_window_procedure);
	SetWindowLong(dialog.rb_mmx_int, GWL_WNDPROC, (LONG)mmx_button_window_procedure);
}

/*-----------------------------------------------------------------*/
static void remove_subclass()
{
	SetWindowLong(dialog.rb_sse2_int, GWL_WNDPROC, (LONG)default_button_window_procedure);
	SetWindowLong(dialog.rb_sse_float, GWL_WNDPROC, (LONG)default_button_window_procedure);
	SetWindowLong(dialog.rb_sse2_double, GWL_WNDPROC, (LONG)default_button_window_procedure);
	SetWindowLong(dialog.rb_mmx_int, GWL_WNDPROC, (LONG)default_button_window_procedure);
}

/*-----------------------------------------------------------------*/
static void enable_simd_radio_button()
{
	SendMessage(dialog.rb_sse2_int, WM_ENABLE, (WPARAM)TRUE, 0);
	SendMessage(dialog.rb_sse_float, WM_ENABLE, (WPARAM)TRUE, 0);
	SendMessage(dialog.rb_sse2_double, WM_ENABLE, (WPARAM)TRUE, 0);
	SendMessage(dialog.rb_mmx_int, WM_ENABLE, (WPARAM)TRUE, 0);
}

/*-----------------------------------------------------------------*/
static void load_ini_value(FILTER *fp)
{
	int set_width;
	int set_height;
	char work[8];
	
	set_width = fp->exfunc->ini_load_int(fp, "width", 640);
	set_height = fp->exfunc->ini_load_int(fp, "height", 480);
		
	if(set_width<track_s[0]){
		set_width=track_s[0];
	}else if(set_width>track_e[0]){
		set_width=track_e[0];
	}
	itoa(set_width, work, 10);
	SetWindowText(dialog.e_width, work);
		
	if(set_height<track_s[1]){
		set_height=track_s[1];
	}else if(set_height>track_e[1]){
		set_height=track_e[1];
	}
	itoa(set_height, work, 10);
	SetWindowText(dialog.e_height, work);

	simd = fp->exfunc->ini_load_int(fp, "simd", 0);
	switch(simd){
	case 0:
		SendMessage(dialog.rb_no_simd, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
		dialog.old_simd = dialog.rb_no_simd;
		fp->func_proc = func_proc;
		break;
	case 1:
		SendMessage(dialog.rb_sse2_int, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
		dialog.old_simd = dialog.rb_sse2_int;
		fp->func_proc = func_proc_sse2_int;
		break;
	case 2:
		SendMessage(dialog.rb_sse_float, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
		dialog.old_simd = dialog.rb_sse_float;
		fp->func_proc = func_proc_sse_float;
		break;
	case 3:
		SendMessage(dialog.rb_sse2_double, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
		dialog.old_simd = dialog.rb_sse2_double;
		fp->func_proc = func_proc_sse2_double;
		break;
	case 4:
		SendMessage(dialog.rb_mmx_int, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
		dialog.old_simd = dialog.rb_mmx_int;
		fp->func_proc = func_proc_mmx_int;
		break;
	}
}

/*-----------------------------------------------------------------*/
static void save_ini_value(FILTER *fp)
{
	char work[8];
	
	GetWindowText(dialog.e_width, work, 5);
	fp->exfunc->ini_save_str(fp, "width", work);
	GetWindowText(dialog.e_height, work, 5);
	fp->exfunc->ini_save_str(fp, "height", work);
	fp->exfunc->ini_save_int(fp, "simd", simd);
}

/*-----------------------------------------------------------------*/
static void on_no_simd_button(FILTER *fp)
{
	SendMessage(dialog.old_simd, BM_SETCHECK, (WPARAM)BST_UNCHECKED, 0);
	SendMessage(dialog.rb_no_simd, WM_KILLFOCUS, 0, 0);
	simd = 0;
	fp->func_proc = func_proc;
	SendMessage(dialog.rb_no_simd, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
	dialog.old_simd = dialog.rb_no_simd;
	
	src_width = 0;
	src_height = 0;
	rst_width = 0;
	rst_height = 0;
}

/*-----------------------------------------------------------------*/
static void on_sse2_int_button(FILTER *fp)
{
	SendMessage(dialog.old_simd, BM_SETCHECK, (WPARAM)BST_UNCHECKED, 0);
	SendMessage(dialog.rb_sse2_int, WM_KILLFOCUS, 0, 0);
	simd = 1;
	fp->func_proc = func_proc_sse2_int;
	SendMessage(dialog.rb_sse2_int, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
	dialog.old_simd = dialog.rb_sse2_int;
	
	src_width = 0;
	src_height = 0;
	rst_width = 0;
	rst_height = 0;
}

/*-----------------------------------------------------------------*/
static void on_sse_float_button(FILTER *fp)
{
	SendMessage(dialog.old_simd, BM_SETCHECK, (WPARAM)BST_UNCHECKED, 0);
	SendMessage(dialog.rb_sse_float, WM_KILLFOCUS, 0, 0);
	simd = 2;
	fp->func_proc = func_proc_sse_float;
	SendMessage(dialog.rb_sse_float, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
	dialog.old_simd = dialog.rb_sse_float;
	
	src_width = 0;
	src_height = 0;
	rst_width = 0;
	rst_height = 0;
}

/*-----------------------------------------------------------------*/
static void on_sse2_double_button(FILTER *fp)
{
	SendMessage(dialog.old_simd, BM_SETCHECK, (WPARAM)BST_UNCHECKED, 0);
	SendMessage(dialog.rb_sse2_double, WM_KILLFOCUS, 0, 0);
	simd = 3;
	fp->func_proc = func_proc_sse2_double;
	SendMessage(dialog.rb_sse2_double, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
	dialog.old_simd = dialog.rb_sse2_double;
	
	src_width = 0;
	src_height = 0;
	rst_width = 0;
	rst_height = 0;
}

/*-----------------------------------------------------------------*/
static void on_mmx_int_button(FILTER *fp)
{
	SendMessage(dialog.old_simd, BM_SETCHECK, (WPARAM)BST_UNCHECKED, 0);
	SendMessage(dialog.rb_mmx_int, WM_KILLFOCUS, 0, 0);
	simd = 4;
	fp->func_proc = func_proc_mmx_int;
	SendMessage(dialog.rb_mmx_int, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
	dialog.old_simd = dialog.rb_mmx_int;
	
	src_width = 0;
	src_height = 0;
	rst_width = 0;
	rst_height = 0;
}

/*-----------------------------------------------------------------*/
static void on_640x480_button(FILTER *fp)
{
	SendMessage(dialog.b640x480, WM_KILLFOCUS,0,0);
	fp->track[0]=640;
	fp->track[1]=480;
	fp->exfunc->filter_window_update(fp);
}

/*-----------------------------------------------------------------*/
static void on_352x240_button(FILTER *fp)
{
	SendMessage(dialog.b352x240, WM_KILLFOCUS,0,0);
	fp->track[0]=352;
	fp->track[1]=240;
	fp->exfunc->filter_window_update(fp);
}

/*-----------------------------------------------------------------*/
static void on_320x240_button(FILTER *fp)
{
	SendMessage(dialog.b320x240, WM_KILLFOCUS,0,0);
	fp->track[0]=320;
	fp->track[1]=240;
	fp->exfunc->filter_window_update(fp);
}

/*-----------------------------------------------------------------*/
static void on_setsize_button(FILTER *fp)
{
	int set_width;
	int set_height;
	char work[8];
	
	SendMessage(dialog.b_set_size, WM_KILLFOCUS,0,0);

	GetWindowText(dialog.e_width, work, 5);
	set_width = atoi(work);
	if(set_width<track_s[0]){
		set_width=track_s[0];
	}else if(set_width>track_e[0]){
		set_width=track_e[0];
	}
			
	GetWindowText(dialog.e_height, work, 5);
	set_height = atoi(work);
	if(set_height<track_s[1]){
		set_height=track_s[1];
	}else if(set_height>track_e[1]){
		set_height=track_e[1];
	}
			
	fp->track[0]=set_width;
	fp->track[1]=set_height;
	fp->exfunc->filter_window_update(fp);
}

/*-----------------------------------------------------------------*/
LRESULT CALLBACK sse_button_window_procedure(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch(msg){
	case WM_ENABLE:
		if(wparam && is_sse_enable()){
			EnableWindow(hwnd, TRUE);
		}else{
			EnableWindow(hwnd, FALSE);
		}
	}

	return CallWindowProc(default_button_window_procedure, hwnd, msg, wparam, lparam);
}

/*-----------------------------------------------------------------*/
LRESULT CALLBACK sse2_button_window_procedure(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch(msg){
	case WM_ENABLE:
		if(wparam && is_sse2_enable()){
			EnableWindow(hwnd, TRUE);
		}else{
			EnableWindow(hwnd, FALSE);
		}
	}

	return CallWindowProc(default_button_window_procedure, hwnd, msg, wparam, lparam);
}

/*-----------------------------------------------------------------*/
LRESULT CALLBACK mmx_button_window_procedure(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch(msg){
	case WM_ENABLE:
		if(wparam && is_mmx_enable()){
			EnableWindow(hwnd, TRUE);
		}else{
			EnableWindow(hwnd, FALSE);
		}
	}

	return CallWindowProc(default_button_window_procedure, hwnd, msg, wparam, lparam);
}

/*-----------------------------------------------------------------*/
LRESULT CALLBACK disable_button_window_procedure(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	switch(msg){
	case WM_ENABLE:
		EnableWindow(hwnd, FALSE);
	}

	return CallWindowProc(default_button_window_procedure, hwnd, msg, wparam, lparam);
}

static void setup_range(FILTER *fp, void *editp)
{
	SYS_INFO si;

	if(fp->exfunc->get_sys_info){
		fp->exfunc->get_sys_info(editp, &si);
		
		track_s[0] = si.min_w;
		track_s[1] = si.min_h;
		track_e[0] = si.max_w;
		track_e[1] = si.max_h;

		fp->exfunc->filter_window_update(fp);
	}
}

